# -*- coding: utf-8 -*-
import asyncio
import hmac
import json
import time

from hashlib import sha1
from qiniu import Auth
from urllib.parse import urlparse
from cachetools import TTLCache

from async_cow import config
from async_cow.compat import b
from async_cow.http.aio import CowClientRequest, CowHttpAuthBase
from async_cow.http.base import RequestBase
from async_cow.service.storage.upload_progress_recorder import UploadProgressRecorder
from async_cow.utils import urlsafe_base64_encode, crc32, _file_iter, rfc_from_timestamp


TTL = 3500
CACHE_MAX_SIZE = 2000


class QiniuAuth(Auth):

    def __init__(self, access_key, secret_key, max_token_level=None):
        """
        :param access_key: access_key
        :param secret_key: secret_key
        :param max_token_level: token缓存数最大水位
        """
        super().__init__(access_key, secret_key)
        
        if not max_token_level:
            max_token_level = CACHE_MAX_SIZE

        self._upload_token_cache = TTLCache(max_token_level, TTL)
        self._rtc_room_token_cache = TTLCache(max_token_level, TTL)

    def get_token(self,
                  bucket,
                  key=None,
                  policy=None,
                  strict_policy=True
                  ):

        token_key = f'{bucket}_{key}_{policy}_{strict_policy}'

        token = self._upload_token_cache.get(token_key, None)

        if not token:

            token = self.upload_token(bucket, key, expires=TTL + 100, policy=policy, strict_policy=strict_policy)
            self._upload_token_cache[token_key] = token

        return token

    def get_rtc_room_token(self, room_access):

        if isinstance(room_access, dict) and 'deadline' in room_access:
            room_access['deadline'] = int(time.time()) + TTL + 100

        token_key = json.dumps(room_access)

        token = self._rtc_room_token_cache.get(token_key, None)

        if not token:

            token = self.token_with_data(token_key)
            self._rtc_room_token_cache[token_key] = token

        return token


class QiniuMacAuth(object):
    """
    Sign Requests

    Attributes:
        _access_key
        _secret_key

    http://kirk-docs.qiniu.com/apidocs/#TOC_325b437b89e8465e62e958cccc25c63f
    """

    def __init__(self, access_key, secret_key):
        self.qiniu_header_prefix = "X-Qiniu-"
        self._access_key = access_key
        self._secret_key = b(secret_key)

    def _token(self, data):
        data = b(data)
        hashed = hmac.new(self._secret_key, data, sha1)
        return urlsafe_base64_encode(hashed.digest())

    def token_of_request(
            self,
            method,
            host,
            url,
            qheaders,
            content_type=None,
            body=None):
        """
        <Method> <PathWithRawQuery>
        Host: <Host>
        Content-Type: <ContentType>
        [<X-Qiniu-*> Headers]

        [<Body>] #这里的 <Body> 只有在 <ContentType> 存在且不为 application/octet-stream 时才签进去。

        """
        parsed_url = urlparse(url)
        netloc = parsed_url.netloc
        path = parsed_url.path
        query = parsed_url.query

        if not host:
            host = netloc

        path_with_query = path
        if query != '':
            path_with_query = ''.join([path_with_query, '?', query])
        data = ''.join(["%s %s" %
                        (method, path_with_query), "\n", "Host: %s" %
                        host, "\n"])

        if content_type:
            data += "Content-Type: %s" % (content_type) + "\n"

        data += qheaders
        data += "\n"

        if content_type and content_type != "application/octet-stream" and body:
            if isinstance(body, bytes):
                data += body.decode(encoding='UTF-8')
            else:
                data += body
        return '{0}:{1}'.format(self._access_key, self._token(data))

    def qiniu_headers(self, headers):
        res = ""
        for key in headers:
            if key.startswith(self.qiniu_header_prefix):
                res += key + ": %s\n" % (headers.get(key))
        return res


class QiniuMacRequestsAuth(CowHttpAuthBase):

    def __call__(self, r: CowClientRequest):
        token = self.auth.token_of_request(
            r.method, r.headers.get('Host', None),
            r.url, self.auth.qiniu_headers(r.headers),
            r.headers.get('Content-Type', None),
            r.body
        )
        r.headers['Authorization'] = 'Qiniu {0}'.format(token)
        return r


class _Resume(object):
    """断点续上传类

    该类主要实现了分块上传，断点续上，以及相应地创建块和创建文件过程，详细规格参考：
    http://developer.qiniu.com/docs/v6/api/reference/up/mkblk.html
    http://developer.qiniu.com/docs/v6/api/reference/up/mkfile.html

    Attributes:
        up_token:                   上传凭证
        key:                        上传文件名
        input_stream:               上传二进制流
        data_size:                  上传流大小
        params:                     自定义变量，规格参考 http://developer.qiniu.com/docs/v6/api/overview/up/response/vars.html#xvar
        mime_type:                  上传数据的mimeType
        progress_handler:           上传进度
        upload_progress_recorder:   记录上传进度，用于断点续传
        modify_time:                上传文件修改日期
        hostscache_dir：            host请求 缓存文件保存位置
    """

    def __init__(self, up_token, key, input_stream, file_name, data_size, hostscache_dir, params, mime_type,
                 progress_handler, upload_progress_recorder, modify_time, keep_last_modified, **settings):
        """初始化断点续上传"""
        self.up_token = up_token
        self.key = key
        self.input_stream = input_stream
        self.file_name = file_name
        self.size = data_size
        self.hostscache_dir = hostscache_dir
        self.params = params
        self.mime_type = mime_type
        self.progress_handler = progress_handler
        self.upload_progress_recorder = upload_progress_recorder or UploadProgressRecorder()
        self.modify_time = modify_time or time.time()
        self.keep_last_modified = keep_last_modified

        if settings.get('http', None):
            self._http = settings.get('http', None)
        else:
            self._http = RequestBase(**settings)

    def record_upload_progress(self, offset):
        record_data = {
            'size': self.size,
            'offset': offset,
            'contexts': [block['ctx'] for block in self.blockStatus]
        }
        if self.modify_time:
            record_data['modify_time'] = self.modify_time

        self.upload_progress_recorder.set_upload_record(self.file_name, self.key, record_data)

    def recovery_from_record(self):
        record = self.upload_progress_recorder.get_upload_record(self.file_name, self.key)
        if not record:
            return 0

        try:
            if not record['modify_time'] or record['size'] != self.size or \
                    record['modify_time'] != self.modify_time:
                return 0
        except KeyError:
            return 0
        self.blockStatus = [{'ctx': ctx} for ctx in record['contexts']]
        return record['offset']

    async def upload(self):
        """上传操作"""
        self.blockStatus = []
        if config.get_default('default_zone').up_host:
            host = config.get_default('default_zone').up_host
        else:
            host = await config.get_default('default_zone').get_up_host_by_token(self.up_token, self.hostscache_dir)

        offset = self.recovery_from_record()
        async for block in _file_iter(self.input_stream, config._BLOCK_SIZE, offset):
            length = len(block)
            crc = crc32(block)
            ret, info = await self.make_block(block, length, host)
            if ret is None and not info.need_retry():
                return ret, info
            if info.connect_failed():
                if config.get_default('default_zone').up_host_backup:
                    host = config.get_default('default_zone').up_host_backup
                else:
                    host = await config.get_default('default_zone').get_up_host_backup_by_token(self.up_token,
                                                                                          self.hostscache_dir)
            if info.need_retry() or crc != ret['crc32']:
                ret, info = await self.make_block(block, length, host)
                if ret is None or crc != ret['crc32']:
                    return ret, info

            self.blockStatus.append(ret)
            offset += length
            self.record_upload_progress(offset)

            if asyncio.iscoroutinefunction(self.progress_handler):
                await self.progress_handler(((len(self.blockStatus) - 1) * config._BLOCK_SIZE) + length, self.size)
            elif (callable(self.progress_handler)):
                self.progress_handler(((len(self.blockStatus) - 1) * config._BLOCK_SIZE) + length, self.size)

        return await self.make_file(host)

    async def make_block(self, block, block_size, host):
        """创建块"""
        url = self.block_url(host, block_size)
        return await self.post(url, block)

    def block_url(self, host, size):
        return '{0}/mkblk/{1}'.format(host, size)

    def file_url(self, host):
        url = ['{0}/mkfile/{1}'.format(host, self.size)]

        if self.mime_type:
            url.append('mimeType/{0}'.format(urlsafe_base64_encode(self.mime_type)))

        if self.key is not None:
            url.append('key/{0}'.format(urlsafe_base64_encode(self.key)))

        if self.file_name is not None:
            url.append('fname/{0}'.format(urlsafe_base64_encode(self.file_name)))

        if self.params:
            for k, v in self.params.items():
                url.append('{0}/{1}'.format(k, urlsafe_base64_encode(v)))
            pass

        if self.modify_time and self.keep_last_modified:
            url.append(
                "x-qn-meta-!Last-Modified/{0}".format(urlsafe_base64_encode(rfc_from_timestamp(self.modify_time))))

        url = '/'.join(url)
        # print url
        return url

    async def make_file(self, host):
        """创建文件"""
        url = self.file_url(host)
        body = ','.join([status['ctx'] for status in self.blockStatus])
        self.upload_progress_recorder.delete_upload_record(self.file_name, self.key)
        return await self.post(url, body)

    async def post(self, url, data):
        return await self._http._post_with_token(url, data, self.up_token)



# -*- coding: utf-8 -*-

import json


class Sms:
    """
    TODO 所有功能方法都将返回协程对象，必须用await调用，和原七牛SDK保持接口同步
    文档：https://developer.qiniu.com/sms/5812/sms-product-introduction
    Attributes:
        cow: AsyncCow对象
        不希望你直接调用本类
        而是通过下列方式调用

        sms = AsyncCow(<access_key>, <secret_key>).get_sms()
        sig_dic = await sms.createSignature(<signature>, <source>, <pics>)
    """

    def __init__(self, cow):
        self.cow = cow
        self.server = 'https://sms.qiniuapi.com'

    def createSignature(self, signature, source, pics=None):
        """
        *创建签名
        *signature: string类型，必填，【长度限制8个字符内】超过长度会报错
        *source: string类型，必填，申请签名时必须指定签名来源。取值范围为：
            enterprises_and_institutions 企事业单位的全称或简称
            website 工信部备案网站的全称或简称
            app APP应用的全称或简称
            public_number_or_small_program 公众号或小程序的全称或简称
            store_name 电商平台店铺名的全称或简称
            trade_name 商标名的全称或简称
        *pics: 签名对应的资质证明图片进行 base64 编码格式转换后的字符串
        * @ return: 类型array
        {
            "signature_id": < signature_id >
        }
        """
        req = {}
        req['signature'] = signature
        req['source'] = source
        if pics:
            req['pics'] = pics
        body = json.dumps(req)
        url = '{0}/v1/signature'.format(self.server)
        return self._post(url, body)

    def querySignature(self, audit_status=None, page=1, page_size=20):
        """
        查询签名
        * audit_status: 审核状态 string 类型，可选，取值范围为: "passed"(通过), "rejected"(未通过), "reviewing"(审核中)
        * page:页码 int  类型，
        * page_size: 分页大小 int 类型，可选， 默认为20
        *@return: 类型array {
            "items": [{
            "id": string,
            "signature": string,
            "source": string,
            "audit_status": string,
            "reject_reason": string,
            "created_at": int64,
            "updated_at": int64
                }...],
            "total": int,
            "page": int,
            "page_size": int,
            }
        """
        url = '{0}/v1/signature'.format(self.server)
        if audit_status:
            url = '{0}?audit_status={1}&page={2}&page_size={3}'.format(url, audit_status, page, page_size)
        else:
            url = '{0}?page={1}&page_size={2}'.format(url, page, page_size)
        return self._get(url)

    def updateSignature(self, id, signature):
        """
        编辑签名
        *  id 签名id : string 类型，必填，
        * signature: string 类型，必填，
        request 类型array {
        "signature": string
        }
        :return:
        """
        url = '{0}/v1/signature/{1}'.format(self.server, id)
        req = {}
        req['signature'] = signature
        body = json.dumps(req)
        return self._put(url, body)

    def deleteSignature(self, id):

        """
        删除辑签名
        *  id 签名id : string 类型，必填，
        * @retrun : 请求成功 HTTP 状态码为 200

        """
        url = '{0}/v1/signature/{1}'.format(self.server, id)
        return self._delete(url)

    def createTemplate(self, name, template, type, description, signature_id):
        """
        创建模版
        :param name: 模板名称 string 类型 ，必填
        :param template: 模板内容 string  类型，必填
        :param type: 模板类型 string 类型，必填，
                    取值范围为: notification (通知类短信), verification (验证码短信), marketing (营销类短信)
        :param description: 申请理由简述 string  类型，必填
        :param signature_id: 已经审核通过的签名 string  类型，必填
        :return: 类型 array {
        "template_id": string
                }
        """
        url = '{0}/v1/template'.format(self.server)
        req = {}
        req['name'] = name
        req['template'] = template
        req['type'] = type
        req['description'] = description
        req['signature_id'] = signature_id
        body = json.dumps(req)
        return self._post(url, body)

    def queryTemplate(self, audit_status, page=1, page_size=20):
        """
        查询模版
        :param audit_status: 审核状态, 取值范围为: passed (通过), rejected (未通过), reviewing (审核中)
        :param page: 页码。默认为 1
        :param page_size: 分页大小。默认为 20
        :return:{
        "items": [{
            "id": string,
            "name": string,
            "template": string,
            "audit_status": string,
            "reject_reason": string,
            "type": string,
            "signature_id": string, // 模版绑定的签名ID
            "signature_text": string, // 模版绑定的签名内容
            "created_at": int64,
            "updated_at": int64
            }...],
            "total": int,
            "page": int,
            "page_size": int
        }
        """
        url = '{0}/v1/template'.format(self.server)
        if audit_status:
            url = '{0}?audit_status={1}&page={2}&page_size={3}'.format(url, audit_status, page, page_size)
        else:
            url = '{0}?page={1}&page_size={2}'.format(url, page, page_size)
        return self._get(url)

    def updateTemplate(self, id, name, template, description, signature_id):
        """
        更新模版
        :param id: template_id
        :param name: 模板名称 string 类型 ，必填
        :param template: 模板内容 string  类型，必填
        :param description: 申请理由简述 string  类型，必填
        :param signature_id: 已经审核通过的签名 string  类型，必填
        :return: 请求成功 HTTP 状态码为 200
        """
        url = '{0}/v1/template/{1}'.format(self.server, id)
        req = {}
        req['name'] = name
        req['template'] = template
        req['description'] = description
        req['signature_id'] = signature_id
        body = json.dumps(req)
        return self._put(url, body)

    def deleteTemplate(self, id):
        """
        删除模版
        :param id: template_id
        :return: 请求成功 HTTP 状态码为 200
        """
        url = '{0}/v1/template/{1}'.format(self.server, id)
        return self._delete(url)

    def sendMessage(self, template_id, mobiles, parameters):
        """
        发送短信
        :param template_id:  模板 ID
        :param mobiles: 手机号
        :param parameters: 自定义魔法变量，变量设置在创建模板时，参数template指定
        :return:{
            "job_id": string
        }
        短信发送给用户后，将会通过回调业务 URL 的方式，通知业务方用户发送短信的状态。
        回调文档见：https://developer.qiniu.com/sms/5910/message-push

        """
        url = '{0}/v1/message'.format(self.server)
        req = {}
        req['template_id'] = template_id
        req['mobiles'] = mobiles
        req['parameters'] = parameters
        body = json.dumps(req)
        return self._post(url, body)

    def get_charge_message_count(self, start, end, g, status):
        """
        查询发送计费条数
        https://developer.qiniu.com/sms/7926/query-send-billing-number
        名称	必填	描述
        start	是	查询开始时间。格式：2006-01-02
        end	是	查询结束时间。格式：2006-01-02
        g	是	计量数据聚合粒度，支持： day, 5min
        status	是	短信的状态，支持：“success”(发送成功), “failed”(发送失败), “sending”(发送中)
        :return  通知短信和验证码短信会统一到 系统短信 计量计费
        {
            "data": {
                "marketing": { // 营销短信统计
                    "times": [
                        1609430400,
                    ],
                    "values": [
                        0,
                    ]
                },
                "notification": { // 通知短信统计
                    "times": [
                        1609430400,
                    ],
                    "values": [
                        0,
                    ]
                },
                "verification": {// 验证码短信统计
                    "times": [
                        1609430400,
                    ],
                    "values": [
                        0,
                    ]
                },
                "voice": { // 语音短信统计
                    "times": [
                        1609430400,
                    ],
                    "values": [
                        0,
                    ]
                }
            }
        }

        """

        url = f'/v1/user/statistics?start={start}&end={end}&g={g}&status={status}'

        return self._get(url)

    def get_messages_info(self,
                          job_id=None,
                          message_id=None,
                          mobile=None,
                          status=None,
                          template_id=None,
                          type=None,
                          start=None,
                          end=None,
                          page=1,
                          page_size=20,
                          ):
        """
        查询发送记录，文档：https://developer.qiniu.com/sms/api/5852/query-send-sms

        名称	必填	描述
        job_id	否	发送任务返回的 id
        message_id	否	单条短信发送接口返回的 id
        mobile	否	接收短信的手机号码
        status	否	短信的状态，sending: 发送中，success: 发送成功，failed: 发送失败，waiting: 等待发送
        template_id	否	模版 id
        type	否	短信类型，marketing: 营销短信，notification: 通知短信，verification: 验证码类短信，voice: 语音短信
        start	否	开始时间，timestamp，例如: 1563280448
        end	否	结束时间，timestamp，例如: 1563280471
        page	否	页码，默认为 1
        page_size	否	每页返回的数据条数，默认20，最大200

        :return:
        {
            "page": int,
            "page_size": int,
            "items":[{
                "message_id": string,
                "job_id": string,
                "mobile": string,
                "content": string,
                "status": string,
                "type": string,
                "error": string, // 短信发送失败详细状态信息
                "count": int, // 发送的短信条数
                "created_at": timestamp,  // 短信发送时间
                "delivrd_at": timestamp // 如果短信发送失败，那么不会返回这个字段
            }...]
        }
        """
        params = {
            'page': page,
            'page_size': page_size,
        }
        if job_id is not None:
            params['job_id'] = job_id
        if message_id is not None:
            params['message_id'] = message_id
        if mobile is not None:
            params['mobile'] = mobile
        if status is not None:
            params['status'] = status
        if template_id is not None:
            params['template_id'] = template_id
        if type is not None:
            params['type'] = type
        if start is not None:
            params['start'] = start
        if end is not None:
            params['end'] = end
        
        url = "{0}/v1/messages".format(self.server)
        return self._get(url, params=params)

    def _post(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return self.cow.http._post_with_qiniu_mac_and_headers(url, data, self.cow.auth, headers)

    def _get(self, url, params=None):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        return self.cow.http._get_with_qiniu_mac_and_headers(url, params, self.cow.auth, headers)

    def _put(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return self.cow.http._put_with_qiniu_mac_and_headers(url, data, self.cow.auth, headers)

    def _delete(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return self.cow.http._delete_with_qiniu_mac_and_headers(url, data, self.cow.auth, headers)
